section \<open>Abstract XOR\<close>
theory "Abstract_XOR"
  imports
    "../infrastructure/Message_set"
begin

text\<open>We model xor as an operation on finite sets (fset). {||} is defined as the identity element.\<close>

text\<open>xor of two fsets is the symmetric difference\<close>
definition xor :: "'a fset \<Rightarrow> 'a fset \<Rightarrow> 'a fset" where
  "xor xs ys = (xs |\<union>| ys) |-| (xs |\<inter>| ys)"

lemma xor_singleton: 
  "xor xs {| z |} = (if z |\<in>| xs then xs |-| {| z |} else finsert z xs)"
  "xor {| z |} xs = (if z |\<in>| xs then xs |-| {| z |} else finsert z xs)"
  by (auto simp add: xor_def)

(*try using safe instead of auto*)

declare finsertCI[rule del]
declare finsertCI[intro]

lemma xor_assoc: "xor (xor xs ys) zs = xor xs (xor ys zs)"
  by (auto simp add: xor_def)

lemma xor_commut: "xor xs ys = xor ys xs"
  by (auto simp add: xor_def)

lemma xor_self_inv: "\<lbrakk>xor xs ys = zs; xs = ys\<rbrakk> \<Longrightarrow> zs = {||}"
  by (auto simp add: xor_def)

lemma xor_self_inv': "xor xs xs = {||}"
  by (auto simp add: xor_def)

lemma xor_identity1[simp]: "xor xs {||} = xs"
  by (auto simp add: xor_def)

lemma xor_identity2[simp]: "xor {||} xs = xs"
  by (auto simp add: xor_def)

lemma xor_in: "z |\<in>| xs \<Longrightarrow> z |\<notin>| (xor xs {| z |})"
  by (auto simp add: xor_singleton)

lemma xor_out: "z |\<notin>| xs \<Longrightarrow> z |\<in>| (xor xs {| z |})"
  by (auto simp add: xor_singleton)

lemma xor_elem1[dest]: "\<lbrakk>x \<in> fset (xor X Y); x |\<notin>| X\<rbrakk> \<Longrightarrow> x |\<in>| Y"
  by(auto simp add: xor_def)

lemma xor_elem2[dest]: "\<lbrakk>x \<in> fset (xor X Y); x |\<notin>| Y\<rbrakk> \<Longrightarrow> x |\<in>| X"
  by(auto simp add: xor_def)


lemma size_fset: "size (FS x) > size x"
  by (auto simp add: sum_Suc)

lemma uinfo_mono: "\<lbrakk>uinfo_s = finsert (f (FS uinfo_s)) uinfo_b; \<And> x. size (f x) > size x\<rbrakk> \<Longrightarrow> False"
  by (metis (no_types) add.right_neutral add_Suc_right finite_fset finsert.rep_eq less_add_Suc1 
      msgterm.size(17) not_less_eq size_fset_simps sum.insert_remove)

lemma uinfo_contains_mac:
  assumes "uinfo_s = xor uinfo_b {| mac |}" "mac = f (FS uinfo_s)" "\<And> x. size (f x) > size x"
  shows "mac |\<in>| uinfo_b"
  using assms(1)
  apply(auto simp add: xor_def)
  using uinfo_mono
  by (metis assms notin_fset xor_singleton)

lemma uinfo_is_finsert_mac:
  assumes "uinfo_s = xor uinfo_b {| mac |}" "mac = f (FS uinfo_s)" "\<And> x. size (f x) > size x"
  shows "uinfo_b = finsert mac uinfo_s"
  using assms uinfo_contains_mac
  by (metis finsert_fminus xor_singleton)

thm uinfo_is_finsert_mac[where ?f="(\<lambda>x. Hash x)"]
thm uinfo_is_finsert_mac[where ?f="(\<lambda>x. Mac[k] (L [Num ts, upif, downif, x]))"]







fun no_singleton_fset :: "'a fset \<Rightarrow> bool" where
  "no_singleton_fset t = (if fcard t = Suc 0 then False else True)"

fun no_singleton :: "msgterm \<Rightarrow> bool" where
  "no_singleton (FS fs) = (no_singleton_fset fs)" 
| "no_singleton t = True"

fun xor_term0 :: "msgterm fset \<Rightarrow> msgterm fset \<Rightarrow> msgterm" where
  "xor_term0 xs ys = 
    (if (\<exists>x. xor xs ys = {| x |})
then (THE x. xor xs ys = {| x |}) 
        else FS (xor xs ys))"

(* {| x |} is finsert x {||}*)

fun xor_term1 :: "msgterm fset \<Rightarrow> msgterm \<Rightarrow> msgterm" where
  "xor_term1 xs (FS ys) = xor_term0 xs ys"
| "xor_term1 xs y = xor_term0 xs {| y |}"

fun xor_term :: "msgterm \<Rightarrow> msgterm \<Rightarrow> msgterm" where
  "xor_term (FS xs) y = xor_term1 xs y"
| "xor_term x y = xor_term1 {| x |} y"

lemma "xor_term (FS xs) (FS ys) = xor_term0 xs ys" by auto


lemma fcard_1[simp]: "fcard xs = Suc 0 \<longleftrightarrow> (\<exists>x. xs = {| x |})" 
  apply auto
  using  Zero_not_Suc all_not_fin_conv fcard_fempty fcard_finsert 
         fcard_fminus_fsingleton_if finsert_absorb finsert_fminus_single fminus_cancel
  apply (metis (full_types))+
  done

lemma fset_singleton_not_empty[dest]: "{||} = {|x|} \<Longrightarrow> False" by auto

lemma 
  "\<lbrakk>\<And>x . x |\<in>| X \<Longrightarrow> no_singleton x; \<And>x . x |\<in>| Y \<Longrightarrow> no_singleton x\<rbrakk> 
  \<Longrightarrow> no_singleton (xor_term0 X Y)" 
  apply auto
  subgoal for x
    apply(cases "x |\<in>| X")
    apply auto 
    by fastforce
  done

lemma "xor_term X Y \<noteq> FS {| x |}" 
(*does not hold! E.g.:
X = FS {| FS {| \<epsilon> |} |}, Y = FS {| |}, then xor_term X Y = FS {| \<epsilon> |}
*)
(*
  apply auto
  apply(cases X, auto)
  subgoal
    apply(cases Y, auto)
    using fset_singleton_not_empty xor_self_inv' msgterm.inject(6)
     apply (metis)
    apply(auto simp add: xor_def)
    subgoal for fs
      apply(cases fs) apply auto
       apply (simp add: xor_identity)

      find_theorems name: "if" name: "split"
      apply (split if_split_asm)
      subgoal  sorry
      by(auto simp add: xor_def)
*)
    oops


declare fminusE[rule del] 
declare finsertCI[rule del]
(*add back the removed rules without !*)
declare fminusE[elim]
declare finsertCI[intro]

lemma fset_dualton: "\<lbrakk>xs |-| {|x|} = {|y|}; x \<in> fset xs\<rbrakk> \<Longrightarrow> xs = {| x, y |}" 
  by auto (*takes a few seconds*)

(*
lemma [simp]: "x \<in> xor xs ys = (if then else)"
*)
lemma xor_is_singleton:
  "\<lbrakk>xor xs ys = {|z|}; z |\<notin>| ys\<rbrakk> \<Longrightarrow> z |\<in>| xs"
  "\<lbrakk>xor xs ys = {|z|}; z |\<notin>| xs\<rbrakk> \<Longrightarrow> z |\<in>| ys"
  by fastforce+

lemma xor_is_singleton':
  "xor xs ys = {|z|} \<Longrightarrow> z |\<notin>| xs \<longrightarrow> z |\<in>| ys \<and> z |\<in>| xs \<longrightarrow> z |\<notin>| ys"
  by blast

lemma normalized_FS: "\<lbrakk>normalized (FS xs); x |\<in>| xs\<rbrakk> \<Longrightarrow> normalized x"
  by(auto simp add: normalized.simps[of "FS xs"])
lemma normalized_FS_FS: "\<lbrakk>normalized (FS xs); x |\<in>| xs\<rbrakk> \<Longrightarrow> x \<noteq> FS ys"
  by(auto simp add: normalized.simps[of "FS xs"])

lemma xor_normalized_singleton: 
  "\<lbrakk>xor xs {|b|} = {|a|}; normalized b; normalized (FS xs)\<rbrakk> \<Longrightarrow> normalized a"
  by (auto elim: normalized_FS dest: xor_is_singleton(1))

lemma xor_normalized_singleton'[elim]: 
  "\<lbrakk>xor xs {|b|} = {|a|}; normalized b; \<And>x . x |\<in>| xs \<Longrightarrow> normalized x\<rbrakk> \<Longrightarrow> normalized a"
  by (auto elim: normalized_FS dest: xor_is_singleton(1))

lemma xor_normalized_nonFS: 
  "\<lbrakk>\<forall>x. xor xs {|b|} \<noteq> {|x|}; normalized b; normalized (FS xs); \<forall>ys. b \<noteq> FS ys\<rbrakk> 
  \<Longrightarrow> normalized (FS (xor xs {|b|}))"
        apply(rule normalized.FSt)
        apply auto 
   apply (metis fempty_iff finsertE normalized_FS notin_fset xor_elem1)
  by (metis finsertE fminusD1 normalized_FS_FS notin_fset xor_singleton(1))

lemma xor_normalized_nonFS'[elim]: 
  "\<lbrakk>\<forall>x. xor xs {|b|} \<noteq> {|x|}; normalized b; \<And>x . x |\<in>| xs \<Longrightarrow> normalized x; \<And>x xs' . x |\<in>| xs \<Longrightarrow> x \<noteq> FS xs'; \<forall>ys. b \<noteq> FS ys\<rbrakk> 
  \<Longrightarrow> normalized (FS (xor xs {|b|}))"
        apply(rule normalized.FSt)
        apply auto 
  subgoal for zs
    apply(cases "FS zs = b") apply auto
    apply(drule xor_elem2) by auto
  done

lemma xor_term1_normalized:
  assumes "\<And>x . x |\<in>| xs \<Longrightarrow> normalized x" "\<And>x xs' . x |\<in>| xs \<Longrightarrow> x \<noteq> FS xs'" "normalized y" 
  shows "normalized (xor_term1 xs y)"
  apply(cases "\<exists>ys . y = FS ys")
    subgoal apply(auto)
      subgoal for ys a
        apply(cases "a |\<in>| xs")
        using assms 
        apply (auto dest: normalized_FS xor_is_singleton(2))
      done
      subgoal
        apply(rule normalized.FSt)
          apply auto 
        subgoal using assms apply auto by (meson normalized_FS notin_fset xor_elem1)
        using assms(2-3) normalized_FS_FS notin_fset xor_elem1 by metis
      done
    apply auto
    apply(cases y) apply auto
    subgoal apply(erule xor_normalized_singleton') using assms by auto 
    subgoal apply(erule xor_normalized_nonFS') using assms by auto
    using assms by (fastforce elim!: xor_normalized_singleton' xor_normalized_nonFS')+

lemma normalized_xor_term: 
  assumes "normalized x" "normalized y" 
  shows "normalized (xor_term x y)"
  apply(cases "\<exists>xs . x = FS xs") apply auto 
  subgoal for xs
    apply(cases "\<exists>ys . y = FS ys") apply(auto)
    subgoal for ys a
      apply(cases "a |\<in>| xs")
      using assms 
      by (auto dest: normalized_FS xor_is_singleton(2))
    subgoal
      using assms apply (auto intro!: normalized.FSt)
      by (meson normalized_FS normalized_FS_FS notin_fset xor_elem1)+
    subgoal
      using assms
      by (auto simp add: normalized_FS intro: xor_term1_normalized dest: normalized_FS_FS)
  done
 subgoal (*xs is not an FS*)
   apply(cases x) using assms by (auto intro!: xor_term1_normalized)
  done

lemma "finsert x {| y |} = finsert y {| x |}"
  by auto

(*ALTERNATIVE: We extend the intruder knowledge by rules that allow him to derive the normal form
from an equivalent non-normal term:

finsert (FS xs) ys \<longrightarrow> xor xs ys      -- (no nesting)
FS {| x |} \<longrightarrow> x                      -- (singleton)

*)

(*COULD BE A USEFUL OVERAPPROXIMATION!*)
lemma "parts {x, y} \<subseteq> parts {xor_term x y}"
  apply(cases "\<exists>xs . x = FS xs") apply auto 
  subgoal for xs
    apply(cases "\<exists>ys . y = FS ys") apply(auto)
    subgoal for ys a
      apply (auto simp add: xor_def)
      oops

lemma xor_term_cancelation: "\<lbrakk>xor_term x y = FS {||}; \<And>fs. x \<noteq> FS fs; \<And>fs. y \<noteq> FS fs\<rbrakk> \<Longrightarrow> x = y"
  apply(cases x) apply auto 
  subgoal by(cases y, auto simp add: xor_def)
  subgoal apply(cases y) apply auto apply(auto simp add: xor_def)

    sorry
  sorry
(*swapped parameters in premise*)
lemma xor_term_cancelation': "\<lbrakk>FS {||} = xor_term x y; \<And>fs. x \<noteq> FS fs; \<And>fs. y \<noteq> FS fs\<rbrakk> \<Longrightarrow> x = y"
  using xor_term_cancelation by simp

lemma xor_term_cancelation_universal: "\<lbrakk>xor_term x y = FS {||}; no_singleton x; no_singleton y\<rbrakk> \<Longrightarrow> x = y"
  sorry

lemma no_singleton_xor_term: "\<lbrakk>t = xor_term x y; no_singleton x; no_singleton y\<rbrakk> \<Longrightarrow> no_singleton t" 
  apply auto
  apply(cases x) apply auto 
  subgoal by(cases y, auto simp add: xor_def)



(* NOT NEEDED FOR NOW

   MAY BE USEFUL TO SHOW THAT normalized IS EQUIVALENT TO normalize t = t. *)













locale abstract_xor 
begin
sublocale comp_fun_commute xor
  by(auto simp add: comp_fun_commute_def xor_def)

(*ffold :: "('a \<Rightarrow> 'b \<Rightarrow> 'b) \<Rightarrow> 'b \<Rightarrow> 'a fset \<Rightarrow> 'b"*)
(*ffold :: "('a fset \<Rightarrow> 'b \<Rightarrow> 'b) \<Rightarrow> 'b \<Rightarrow> 'a fset fset \<Rightarrow> 'b"*)

definition Xor :: "'a fset fset \<Rightarrow> 'a fset" where 
  "Xor xss = ffold xor {||} xss"


fun make_fset :: "msgterm \<Rightarrow> msgterm fset" where
  "make_fset (FS xs) = xs"
| "make_fset x = {|x|}"


fun normalize_fset :: "msgterm fset \<Rightarrow> msgterm fset" where
  "normalize_fset xs = Xor (make_fset |`| xs)"

lemma fset_size_le[intro]:
  assumes "x \<in> fset xs" 
  shows "size x < Suc (\<Sum>x\<in>fset xs. Suc (size x))"
proof-
  have "size x \<le> (\<Sum>x\<in>fset xs. size x)" using assms 
    by (auto intro: member_le_sum)
  moreover have "(\<Sum>x\<in>fset xs. size x) < (\<Sum>x\<in>fset xs. Suc (size x))"
    by (metis assms empty_iff finite_fset lessI sum_strict_mono)
  ultimately show ?thesis by auto
qed

fun normalize :: "msgterm \<Rightarrow> msgterm" where
  "normalize (FS xs)     = FS (normalize_fset (normalize |`| xs))"
| "normalize (L xs)      = L (map normalize xs)"
| "normalize (MPair x y) = MPair (normalize x) (normalize y)"
| "normalize (Hash x)    = Hash (normalize x)"
| "normalize (Crypt x y) = Crypt x (normalize y)"
| "normalize x           = x"

(*TODO: define this inductively, not via normalize!
fun normalized :: "msgterm \<Rightarrow> bool" where
  "normalized t \<longleftrightarrow> (normalize t = t)"
*)

(*to show that we are not restricting the attacker*)
lemma "\<lbrakk>\<And>t. t \<in> H \<Longrightarrow> normalized t\<rbrakk> \<Longrightarrow> normalize ` (synth H) = norm_synth H"
  oops
(* Loops with those two rules with bang
thm finsertCI fminusE

lemma "\<lbrakk>xs' = xs |-| {|z|}\<rbrakk> \<Longrightarrow> f xs xs'"
  apply auto
*)

(*
lemma "\<lbrakk>normalized t; t = FS xs; t' = FS xs'; xs' = xor xs {| z |}; z = Hash z'\<rbrakk> \<Longrightarrow> normalized t'"
  apply(cases "z |\<in>| xs")
  subgoal apply auto
    apply simp
*)
(*
thm fsubsetI
declare fsubsetI[rule del]
declare fsubsetI[intro]

lemma "\<lbrakk>normalized (FS xs); xs = finsert x xs'\<rbrakk> \<Longrightarrow> normalized (FS xs')"
  apply(auto)
  (*
  by (smt abstract_xor.make_fset.simps(1) finsert.rep_eq insert_iff msgterm.distinct(11) msgterm.distinct(27) msgterm.distinct(41) msgterm.distinct(53) msgterm.distinct(63) msgterm.distinct(71) msgterm.distinct(79) msgterm.distinct(81) msgterm.distinct(83) normalized.simps)
 
*) oops

lemma "\<lbrakk>normalized t; t = FS xs; t' = FS xs'; xs = finsert x xs'; 
(make_fset (normalize x)) \<notin> ((\<lambda>x. make_fset (normalize x)) ` fset xs')\<rbrakk> \<Longrightarrow> normalized t'"
  apply(auto)
  apply (auto simp add: ffold_def Finite_Set.fold_def)
  find_theorems name: fold_insert
  oops
lemma "\<lbrakk>normalized t; t = FS xs; t' = FS xs'; xs = finsert x xs'\<rbrakk> \<Longrightarrow> normalized x"
  apply(induction xs)
  apply auto
  apply (auto simp add: ffold_def Finite_Set.fold_def) oops
  find_theorems name: ffold

lemma normalized_subset: "\<lbrakk>normalized t; t = FS xs; t' = FS xs'; xs' |\<subseteq>| xs\<rbrakk> \<Longrightarrow> normalized t'"
  apply auto oops

lemma "\<lbrakk>normalized (FS t); z = Hash z'; normalized z\<rbrakk>
\<Longrightarrow> normalized (FS (xor t {| z |}))"
  apply(cases "z |\<in>| t")
  subgoal 
    apply auto 
    sorry
  apply auto
(*
  apply(induction t)
  apply auto subgoal oops
*)
lemma "\<lbrakk>normalized t; (case t of FS y \<Rightarrow> f (FS y) z = FS (xor y {| z |}) | y \<Rightarrow> f y z = y);
z = Hash z'; normalized z\<rbrakk>
\<Longrightarrow> normalized (f t z)"
  apply auto
  apply(cases t)
           apply auto
  oops

lemma "\<lbrakk>normalized t; \<And> x z . (case x of FS y \<Rightarrow> f (FS y) z = FS (xor y {| z |}) | y \<Rightarrow> f y z = y);
z = Hash z'; normalized z\<rbrakk>
\<Longrightarrow> normalized (f t z)"
  apply auto
  apply(cases t)
  apply auto


abbreviation test_term where
  "test_term x \<equiv> FS {| FS {|x|}|}"


lemma "\<lbrakk>normalized (FS foo); normalized baz\<rbrakk> \<Longrightarrow> normalized (FS (xor foo {| Hash baz |}))"
  apply simp
  apply auto
  oops

lemma "normalize (test_term (Num 0)) = FS {| Num 0 |}"
  by(auto)

lemma "normalize (test_term (L [test_term (Num 0)])) = FS {| L [FS {|Num 0|}] |}"
  by(auto)



find_consts "('a \<Rightarrow> bool) \<Rightarrow> 'a fset"

(*stuff that may be relevant later*)

term sorted_list_of_fset (*this could be useful later*)

(*pick an element from an fset (requires the fset to be non-empty)*)
fun pick :: "'a fset \<Rightarrow> 'a" where
  "pick xs = (SOME x . x \<in> fset xs)"

*)
end
end